#include "opencv2/opencv.hpp"
#include "opencv2/core/ocl.hpp"
#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "time.h"

#define NUM_FRAMES 200

using namespace cv;
using namespace std;


/// Global variables
string window_name = "Edge Map";


UMat* GaussianSmoothening(UMat* I, float Sn) {
    // Gaussian Smoothening
    UMat* IBlur = new UMat(I->rows, I->cols, I->type());
    int kernelSize = 2*ceil(3*Sn)+1;
    GaussianBlur(*I, *IBlur, Size(kernelSize, kernelSize), Sn);
    return IBlur;
}

UMat* NonLinearLaplacian(UMat* IBlur, Mat B) {
    UMat IErode, IDilate;
    UMat* L = new UMat(IBlur->rows, IBlur->cols, IBlur->type());
    erode(*IBlur, IErode, B);
    dilate(*IBlur, IDilate, B);
    //*L = IErode.getMat(ACCESS_READ) + IDilate.getMat(ACCESS_READ) - (2*IBlur->getMat(ACCESS_READ));
    add(IErode, IDilate, IErode);
    add(*IBlur, *IBlur, *IBlur);
    subtract(IErode, *IBlur, *L);
    IErode.release();
    IDilate.release();
    //UMat* L = new UMat(IBlur->rows, IBlur->cols, IBlur->type());
    //Laplacian(*IBlur, *L, -1);
    return L;
}

UMat* ZeroCrossings(UMat* I, Mat SE) {
    UMat PErode, PDilate;
    //Mat P = (*I >= 0);
    UMat P;
    compare(*I, 0, P, CMP_GE);
    UMat* Z = new UMat(I->rows, I->cols, I->type());

    erode(P, PErode, SE);
    dilate(P, PDilate, SE);


    //*Z = PDilate.getMat(ACCESS_READ)-PErode.getMat(ACCESS_READ);
    subtract(PDilate, PErode, *Z);
    P.release();
    PDilate.release();
    PErode.release();
    return Z;
}

UMat* Gradient(UMat* I) {

    UMat dIx, dIy;
    UMat* dI = new UMat(I->rows, I->cols, I->type());
    Sobel(*I, dIx, -1, 1, 0, 1);
    Sobel(*I, dIy, -1, 0, 1, 1);
    magnitude(dIx, dIy, *dI);
    dIx.release();
    dIy.release();
    return dI;
}

UMat* RejectZeroCrossings(UMat* dI, UMat* Z, float Threshold) {
    double dI_max;
    UMat* E = new UMat(Z->rows, Z->cols, Z->type());
    minMaxLoc(*dI, NULL, &dI_max);
    //minMaxLoc(dI->getMat(ACCESS_READ), NULL, &dI_max);
    UMat temp;
    compare(*dI, Threshold*dI_max, temp, CMP_GT);
    bitwise_and(*Z, temp, *E);
    temp.release();
    //*E = Z->getMat(ACCESS_READ) & temp.getMat(ACCESS_READ);
    return E;
}

/* Edge Detect
 * Returns edges of image I in binary matrix E
 * Sn is the standard deviation of Gaussians needed for the filters
 * theta is the threshold used to determine whether there is an edge
 */
UMat* EdgeDetect(UMat* I, float Sn, float Threshold, Mat B) {

    // Gaussian Smoothening
    UMat* IBlur = GaussianSmoothening(I, Sn);
    //cout << "Show GS\n";
    //imshow(window_name, *IBlur);
    //waitKey(0);

    // Gradient Computation
    UMat* dI = Gradient(I);
    //cout << "Show G\n";
    //imshow(window_name, *dI);
    //waitKey(0);

    I->release();
    // Non-Linear Laplacian Estimate
    UMat* L = NonLinearLaplacian(IBlur, B);
    //Mat* L = new Mat(IBlur->rows, IBlur->cols, IBlur->type());
    //Laplacian(*IBlur, *L, -1);
    //cout << "Show L\n";
    //imshow(window_name, *L);
    //waitKey(0);
    IBlur->release();

    // Find zero-crossings
    UMat* Z = ZeroCrossings(L, B);
    //cout << "Show ZC\n";
    //imshow(window_name, *Z);
    //waitKey(0);
    L->release();

    // Reject Zero Crossings
    UMat* E = RejectZeroCrossings(dI, Z, Threshold);
    //cout << "Show RZC\n";
    //imshow(window_name, *E);
    //waitKey(0);
    Z->release();
    dI->release();
    return E;
    //imshow(window_name, E);
    //waitKey(0);
}


/** @function main */
int main( int argc, char** argv )
{
    cout << "Accelerating Using OpenCV" << CV_VERSION << "\n";
    ocl::setUseOpenCL(false);
    cout << "OpenCL: " << ocl::useOpenCL() << "\n";
    /// Load an image
    Mat src = imread( argv[1] );

    if( !src.data )
    {
        return -1;
    }

    /// Convert the image to grayscale
    cvtColor( src, src, CV_BGR2GRAY );

    /// Create a window
    namedWindow( window_name, CV_WINDOW_AUTOSIZE );

    src.convertTo(src, CV_32F, 1.0/255.0);
    Mat B = getStructuringElement(MORPH_CROSS, Size(3,3));
    //src.create(10000, 12000, CV_32FC1);
    Timestamp start = get_time(TIMER_MT);
    for(int i=0; i<NUM_FRAMES; i++) {
        UMat srcUMat = src.getUMat(ACCESS_READ);
        UMat* E = EdgeDetect(&srcUMat, 1.0, 0.1, B);
        //imshow(window_name, *E);
        //waitKey(0);
        E->getMat(ACCESS_RW);
        E->release();
    }

    Timestamp end = get_time(TIMER_MT);
    cout << "Running time  = " << (end-start * 1.0)/BILLION  << " seconds\n";
    /// Wait until user exit program by pressing a key
    //waitKey(0);

    return 0;
}
