/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/* 
 * Kernel of dense matrix-matrix multiplication kernel.
 */

__kernel void mysgemmNT( __global const float *A, int lda, __global const float *B, int ldb, __global float* C, int ldc, int k, float alpha, float beta )
{
/*
    // Partial results 
    float c[8];
    for (int i=0; i < 8; i++)
	c[i] = 0.0f;
    float a[8];
    float b[8];

    int m = get_global_id(0) * 8;
    int n = get_global_id(1);

    for (int i = 0; i < k; ++i) {
        for (int id = 0; id < 8; id++) {
	    a[id] = A[m + id + i * lda];
	    b[id] = B[n + i * ldb];
	    c[id] += a[id] * b[id];
        }
    }

    for (int id = 0; id < 8; id++)
        C[m+id+n*ldc] = C[m+id+n*ldc] * beta + alpha * c[id];
*/

    // Partial results 
    float8 cp = (float8)(0.0f);

    int m = get_global_id(0) * 8;
    int n = get_global_id(1);

    for (int i = 0; i < k; ++i) {
        float8 a = vload8(0, A + (m + i * lda));
        float8 b = (float8)(B[n + i * ldb]);
        cp += a * b;
    }

    float8 c = vload8(0, C + (m+n*ldc));
    c = c * beta + alpha * cp;
    vstore8(c, 0, C + (m+n*ldc));
}
