/***************************************************************************
 *cr
 *cr            (C) Copyright 2010 The Board of Trustees of the
 *cr                        University of Illinois
 *cr                         All Rights Reserved
 *cr
 ***************************************************************************/

/* 
 * Kernel of dense matrix-matrix multiplication kernel.
 */

// Parameters of tile sizes
#define TILE_N 8
#define TILE_TB_HEIGHT 8
#define TILE_M (TILE_N*TILE_TB_HEIGHT)

__kernel void mysgemmNT( __global const float *A, int lda, __global const float *B, int ldb, __global float* C, int ldc, int k, float alpha, float beta )
{

    float c[TILE_N];
    for (int i=0; i < TILE_N; i++)
	c[i] = 0.0f;
   
    int mid = get_local_id(1)*get_local_size(0)+get_local_id(0);
    int m = get_group_id(0) * TILE_M + mid;

    int b_base = 0;

    for (int i = 0; i < k; i+=TILE_TB_HEIGHT) {
	float a; 
        b_base = get_group_id(1) * TILE_N + i * ldb;

	for (int j = 0; j < TILE_TB_HEIGHT; j++) {
	    a = A[m + (i+j)*lda];
	    for (int kk = 0; kk < TILE_N; kk++)
		c[kk] += a * B[b_base + j * ldb + kk];

	}
    }
    int t = ldc * get_group_id(1) * TILE_N + m;
    for (int i = 0; i < TILE_N; i++) {
	C[t+i*ldc] = C[t+i*ldc] * beta + alpha * c[i];
    }
/*
    Will be substituted by this kernel at the llvm level

    // Partial results 
    floatn cp = (floatn)(0.0f);

    int mid = get_local_id(1)*get_local_size(0)+get_local_id(0);
    int m = get_group_id(0) * TILE_M + mid;

    int b_base = 0;

    for (int i = 0; i < k; i+=TILE_TB_HEIGHT) {
	float a;
        b_base = get_group_id(1) * TILE_N + i * ldb;

	for (int j = 0; j < TILE_TB_HEIGHT; j++) {
	    a = A[m + (i+j)*lda];
	    cp += a * vloadn(0, B + b_base + j * ldb);
	}
    }

    cp = alpha * cp;
    float c[TILE_N];
    c[0] = cp.s0;
    c[1] = cp.s1;
    c[2] = cp.s2;
    c[3] = cp.s3;
    c[4] = cp.s4;
    c[5] = cp.s5;
    c[6] = cp.s6;
    c[7] = cp.s7;

    int t = ldc * get_group_id(1) * TILE_N + m;
    for (int i = 0; i < TILE_N; i++) {
	C[t+i*ldc] = C[t+i*ldc] * beta + c[i];
    }

*/

*/

}
