# (c) 2007 The Board of Trustees of the University of Illinois.

# Default language wide options

LANG_CFLAGS=-I$(PARBOIL_ROOT)/common/include -I$(OPENCL_PATH)/include
LANG_CXXFLAGS=$(LANG_CFLAGS)
LANG_LDFLAGS=-lOpenCL -L$(OPENCL_LIB_PATH) -lrt

CFLAGS=$(LANG_CFLAGS) $(PLATFORM_CFLAGS) $(APP_CFLAGS)
CXXFLAGS=$(LANG_CXXFLAGS) $(PLATFORM_CXXFLAGS) $(APP_CXXFLAGS)
LDFLAGS=$(LANG_LDFLAGS) $(PLATFORM_LDFLAGS) $(APP_LDFLAGS)


LLVM_INSTALL:=$(LLVM_SRC_ROOT)/../build
LIBCLC:=$(LLVM_SRC_ROOT)/../../libclc-install
LLVM_CC:=$(LLVM_INSTALL)/bin/clang
LLVM_LINK:=$(LLVM_INSTALL)/bin/llvm-link


# Rules common to all makefiles

########################################
# Functions
########################################

# Add BUILDDIR as a prefix to each element of $1
INBUILDDIR=$(addprefix $(BUILDDIR)/,$(1))

# Add SRCDIR as a prefix to each element of $1
INSRCDIR=$(addprefix $(SRCDIR)/,$(1))


########################################
# Environment variable check
########################################

# The second-last directory in the $(BUILDDIR) path
# must have the name "build".  This reduces the risk of terrible
# accidents if paths are not set up correctly.
ifeq ("$(notdir $(BUILDDIR))", "")
$(error $$BUILDDIR is not set correctly)
endif

ifneq ("$(notdir $(patsubst %/,%,$(dir $(BUILDDIR))))", "build")
$(error $$BUILDDIR is not set correctly)
endif

.PHONY: run
.PRECIOUS: $(BUILDDIR)/%.ll

ifeq ($(OPENCL_PATH),)
FAILSAFE=no_opencl
else 
FAILSAFE=
endif

########################################
# Derived variables
########################################

OBJS = $(call INBUILDDIR,$(SRCDIR_OBJS))
KERNEL = $(call INBUILDDIR,$(KERNEL_OBJS))

ifeq ($(DEBUGGER),)
DEBUGGER=gdb
endif

########################################
# Rules
########################################

default: $(FAILSAFE) $(BUILDDIR) $(BIN) $(KERNEL)

run : $(RUNDIR)
	echo "Resolving OpenCL library..."
	$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) ldd ./$(BIN) | grep OpenCL
	$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) ./$(BIN) $(ARGS)
	$(TOOL) $(OUTPUT) $(REF_OUTPUT)

debug:
	@echo "Resolving OpenCL library..."
	@$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) ldd $(BIN) | grep OpenCL
	@$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) $(DEBUGGER) --args $(BIN) $(ARGS)

clean :
	rm -f $(BUILDDIR)/*
	if [ -a $(BIN) ]; then rm $(BIN); fi
	if [ -d $(BUILDDIR) ]; then rm -rf $(BUILDDIR); fi
	if [ -d $(RUNDIR) ]; then rm -rf $(RUNDIR); fi

$(BIN) : $(OBJS) $(BUILDDIR)/parboil_opencl.o 
	$(CXX) $^ -o $@ $(LDFLAGS)

$(RUNDIR) :
	mkdir -p $(RUNDIR)

$(BUILDDIR) :
	mkdir -p $(BUILDDIR)

$(BUILDDIR)/%.o : $(SRCDIR)/%.c
	$(CC) $(CFLAGS) -c $< -o $@

$(BUILDDIR)/parboil_opencl.o : $(PARBOIL_ROOT)/common/src/parboil_opencl.c
	$(CC) $(CFLAGS) -c $< -o $@

$(BUILDDIR)/%.o : $(SRCDIR)/%.cc
	$(CXX) $(CXXFLAGS) -c $< -o $@

$(BUILDDIR)/%.o : $(SRCDIR)/%.cpp
	$(CXX) $(CXXFLAGS) -c $< -o $@

$(BUILDDIR)/%.nvptx.s : $(BUILDDIR)/%.linked.bc
	$(LLVM_CC) -O3 -target nvptx64-nvidia-nvcl $< -S -o $@

$(BUILDDIR)/%.linked.bc : $(BUILDDIR)/%.ll
	$(LLVM_LINK) $(LIBCLC)/lib/clc/nvptx64--nvidiacl.bc $< -o $@

$(BUILDDIR)/%.ll : $(SRCDIR)/%.cl
	$(LLVM_CC) $(CFLAGS) -Dcl_clang_storage_class_specifiers -isystem $(LIBCLC)/include -include clc/clc.h -target nvptx64-nvidia-nvcl $< -O3 -emit-llvm -S -o $@

$(BUILDDIR)/%.ir : $(SRCDIR)/%.cl
	cd $(SRCDIR); ioc64 -input=kernel.cl -ir=kernel.ir
	cp $(SRCDIR)/kernel.ir $@

no_opencl:
	@echo "OPENCL_PATH is not set. Open $(PARBOIL_ROOT)/common/Makefile.conf to set default value."
	@echo "You may use $(PLATFORM_MK) if you want a platform specific configurations."
	@exit 1

