# (c) 2007 The Board of Trustees of the University of Illinois.

# Default language wide options
LANG_CFLAGS=-I$(PARBOIL_ROOT)/common/include
LANG_CXXFLAGS=$(LANG_CFLAGS) -fno-exceptions
LANG_LDFLAGS=-lOpenCL -L$(OPENCL_LIB_PATH) -lrt

CFLAGS=$(LANG_CFLAGS) $(PLATFORM_CFLAGS) $(APP_CFLAGS)
CXXFLAGS=$(LANG_CXXFLAGS) $(PLATFORM_CXXFLAGS) $(APP_CXXFLAGS)
LDFLAGS=$(LANG_LDFLAGS) $(PLATFORM_LDFLAGS) $(APP_LDFLAGS)

# VISC
LIBCLC_LIB_PATH = $(LLVM_SRC_ROOT)/../libclc/built_libs
#VISC_RT_PATH = $(LLVM_SRC_ROOT)/../build/projects/visc-rt
VISC_RT_PATH = $(LLVM_SRC_ROOT)/tools/hpvm/visc-rt

VISC_RT_LIB = $(VISC_RT_PATH)/visc-rt.ll
#LIBCLC_NVPTX_LIB = $(LIBCLC_LIB_PATH)/nvptx--nvidiacl.bc
LIBCLC_NVPTX_LIB = $(LIBCLC_LIB_PATH)/nvptx64--nvidiacl.bc
#LIBCLC_NVPTX_LIB = nvptx64--nvidiacl.bc

LLVM_34_AS = /opt/llvm/bin/llvm-as

TESTGEN_OPTFLAGS = -load LLVMGenVISC.so -genvisc -globaldce
KERNEL_GEN_FLAGS = -O3 -target nvptx64-nvidia-nvcl

ifeq ($(TARGET),x86)
  DEVICE = SPIR_TARGET
  VISC_OPTFLAGS = -load LLVMBuildDFG.so -load LLVMLocalMem.so -load LLVMDFG2LLVM_SPIR.so -load LLVMDFG2LLVM_X86.so -load LLVMClearDFG.so -localmem -dfg2llvm-spir -dfg2llvm-x86 -clearDFG
  CFLAGS += -DOPENCL_CPU
else ifeq ($(TARGET),seq)
  DEVICE = CPU_TARGET
  VISC_OPTFLAGS = -load LLVMBuildDFG.so -load LLVMDFG2LLVM_X86.so -load LLVMClearDFG.so -dfg2llvm-x86 -clearDFG
else ifeq ($(TARGET),seqx86)
  DEVICE = CPU_OR_SPIR_TARGET
  VISC_OPTFLAGS = -load LLVMBuildDFG.so -load LLVMLocalMem.so -load LLVMDFG2LLVM_SPIR.so -load LLVMDFG2LLVM_X86.so -load LLVMClearDFG.so -localmem -dfg2llvm-spir -dfg2llvm-x86 -clearDFG
  CFLAGS += -DOPENCL_CPU
else ifeq ($(TARGET),seqgpu)
  DEVICE = CPU_OR_GPU_TARGET
  VISC_OPTFLAGS = -load LLVMBuildDFG.so -load LLVMLocalMem.so -load LLVMDFG2LLVM_NVPTX.so -load LLVMDFG2LLVM_X86.so -load LLVMClearDFG.so -localmem -dfg2llvm-nvptx -dfg2llvm-x86 -clearDFG
else
  DEVICE = GPU_TARGET
  VISC_OPTFLAGS = -load LLVMBuildDFG.so -load LLVMLocalMem.so -load LLVMDFG2LLVM_NVPTX.so -load LLVMDFG2LLVM_X86.so -load LLVMClearDFG.so -localmem -dfg2llvm-nvptx -dfg2llvm-x86 -clearDFG
endif

CFLAGS += -DDEVICE=$(DEVICE)
CXXFLAGS += -DDEVICE=$(DEVICE)

HOST_LINKFLAGS =

ifeq ($(TIMER),x86)
  VISC_OPTFLAGS += -visc-timers-x86
else ifeq ($(TIMER),ptx)
  VISC_OPTFLAGS += -visc-timers-ptx
else ifeq ($(TIMER),gen)
  TESTGEN_OPTFLAGS += -visc-timers-gen
else ifeq ($(TIMER),spir)
  TESTGEN_OPTFLAGS += -visc-timers-spir
else ifeq ($(TIMER),no)
else
  ifeq ($(TARGET),x86)
    VISC_OPTFLAGS += -visc-timers-x86 -visc-timers-spir
  else ifeq ($(TARGET),seq)
    VISC_OPTFLAGS += -visc-timers-x86
  else ifeq ($(TARGET),seqx86)
    VISC_OPTFLAGS += -visc-timers-x86 -visc-timers-spir
  else ifeq ($(TARGET),seqgpu)
    VISC_OPTFLAGS += -visc-timers-x86 -visc-timers-ptx
  else
    VISC_OPTFLAGS += -visc-timers-x86 -visc-timers-ptx
  endif
  TESTGEN_OPTFLAGS += -visc-timers-gen
endif

ifeq ($(DABSTRACTION),true)
  VISC_OPTFLAGS += -visc-eda
endif

# Rules common to all makefiles

########################################
# Functions
########################################

# Add BUILDDIR as a prefix to each element of $1
INBUILDDIR=$(addprefix $(BUILDDIR)/,$(1))

# Add SRCDIR as a prefix to each element of $1
INSRCDIR=$(addprefix $(SRCDIR)/,$(1))

PYTHON_LLVM_40_34 = ../llvm-40-34.py

########################################
# Environment variable check
########################################

OPENCV_LIB_PATH=${OpenCV_DIR}/lib

# The second-last directory in the $(BUILDDIR) path
# must have the name "build".  This reduces the risk of terrible
# accidents if paths are not set up correctly.
ifeq ("$(notdir $(BUILDDIR))", "")
$(error $$BUILDDIR is not set correctly)
endif

ifneq ("$(notdir $(patsubst %/,%,$(dir $(BUILDDIR))))", "build")
$(error $$BUILDDIR is not set correctly)
endif

.PHONY: run
.PRECIOUS: $(BUILDDIR)/%.ll

ifeq ($(OPENCL_PATH),)
FAILSAFE=no_opencl
else 
FAILSAFE=
endif

########################################
# Derived variables
########################################

OBJS = $(call INBUILDDIR,$(SRCDIR_OBJS))
TEST_OBJS = $(call INBUILDDIR,$(VISC_OBJS))
PARBOIL_OBJS = $(call INBUILDDIR,parboil.ll)
KERNEL = $(TEST_OBJS).kernels.ll
KERNEL_OPT = $(BUILDDIR)/$(APP).kernels.opt.ll
ifeq ($(TARGET),x86)
  SPIR_ASSEMBLY = $(TEST_OBJS).kernels.bc
else ifeq ($(TARGET),seq)
else ifeq ($(TARGET),seqx86)
  SPIR_ASSEMBLY = $(TEST_OBJS).kernels.bc
else ifeq ($(TARGET),seqgpu)
  KERNEL_LINKED = $(BUILDDIR)/$(APP).kernels.linked.ll
  #KERNEL = $(TEST_OBJS).kernels.ll
  PTX_ASSEMBLY = $(TEST_OBJS).nvptx.s
else
  KERNEL_LINKED = $(BUILDDIR)/$(APP).kernels.linked.ll
  #KERNEL = $(TEST_OBJS).kernels.ll
  PTX_ASSEMBLY = $(TEST_OBJS).nvptx.s
endif
HOST_LINKED = $(BUILDDIR)/$(APP).linked.ll
HOST = $(BUILDDIR)/$(APP).host.ll

ifeq ($(DEBUGGER),)
DEBUGGER=gdb
endif

########################################
# Rules
########################################

default: $(FAILSAFE) $(BUILDDIR) $(PTX_ASSEMBLY) $(SPIR_ASSEMBLY) $(BIN)
#default: $(FAILSAFE) $(BUILDDIR) $(BIN)

run : $(RUNDIR)
	echo "Resolving OpenCL library..."
	$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) ldd ./$(BIN) | grep OpenCL
	#echo "Resolving OpenCV library..."
	#$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCV_LIB_PATH) ldd ./$(BIN) | grep opencv
	$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH):$(OPENCV_LIB_PATH) ./$(BIN) $(ARGS)
	$(TOOL) $(OUTPUT) $(REF_OUTPUT)

debug:
	@echo "Resolving OpenCL library..."
	@$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) ldd $(BIN) | grep OpenCL
	@$(shell echo $(RUNTIME_ENV)) LD_LIBRARY_PATH=$(OPENCL_LIB_PATH) $(DEBUGGER) --args $(BIN) $(ARGS)

clean :
	rm -rf $(BUILDDIR)/*
	if [ -a $(BIN) ]; then rm $(BIN); fi
	if [ -a DataflowGraph.dot ]; then rm DataflowGraph.dot*; fi
	if [ -d $(BUILDDIR) ]; then rm -rf $(BUILDDIR); fi
	if [ -d $(RUNDIR) ]; then rm -rf $(RUNDIR); fi

#$(APP_BINS) : $(PTX_ASSEMBLY) $(BIN)
	#echo Generating $(APP_BINS) ...

#TODO: Fix this. Remove echo to actually compile kernel to ptx
$(PTX_ASSEMBLY) : $(KERNEL_LINKED)
	$(CC) $(KERNEL_GEN_FLAGS) -S $< -o $@

$(KERNEL_LINKED) : $(KERNEL_OPT)
	$(LLVM_LINK) $(LIBCLC_NVPTX_LIB) -S $< -o $@

$(SPIR_ASSEMBLY) : $(KERNEL_OPT)
	python $(PYTHON_LLVM_40_34) $< $(BUILDDIR)/kernel_34.ll
	$(LLVM_34_AS) $(BUILDDIR)/kernel_34.ll -o $@

$(KERNEL_OPT) : $(KERNEL)
	$(OPT) $(APP_OPTFLAGS) -S $< -o $@

$(BIN) : $(HOST_LINKED)
	$(CXX) -O3 $(LDFLAGS) $< -o $@

$(HOST_LINKED) : $(HOST) $(OBJS) $(BUILDDIR)/parboil.ll $(VISC_RT_LIB)
	$(LLVM_LINK) $^ -S -o $@

$(VISC_RT_LIB) : $(VISC_RT_PATH)/visc-rt.cpp
	make -C $(LLVM_LIB_PATH)

$(HOST) $(KERNEL): $(BUILDDIR)/$(VISC_OBJS)
	$(OPT) $(VISC_OPTFLAGS) -S $< -o $(HOST)

$(RUNDIR) :
	mkdir -p $(RUNDIR)

$(BUILDDIR) :
	mkdir -p $(BUILDDIR)

$(BUILDDIR)/%.ll: $(SRCDIR)/%.c
	$(CC) $(CFLAGS) -S -emit-llvm $< -o $@

$(BUILDDIR)/%.ll : $(SRCDIR)/%.cc
	$(CXX) $(CXXFLAGS) -S -emit-llvm $< -o $@

$(BUILDDIR)/%.ll : $(SRCDIR)/%.cpp
	$(CXX) $(CXXFLAGS) -S -emit-llvm $< -o $@

$(BUILDDIR)/%.visc.ll: $(BUILDDIR)/%.ll
	$(OPT) $(TESTGEN_OPTFLAGS) $< -S -o $@
	cat $(LLVM_SRC_ROOT)/tools/hpvm/test/parboil/RUN.parboil.script $@ > $@.tmp
	mv $@.tmp $(BUILDDIR)/$(APP).visc.ll
	#@cp $(VISC_OBJS) $(BUILDDIR)/$(VISC_OBJS)

$(BUILDDIR)/%.o : $(SRCDIR)/%.c
	$(CC) $(CFLAGS) -c $< -o $@

$(BUILDDIR)/parboil.ll : $(PARBOIL_ROOT)/common/src/parboil.c
	$(CC) $(CFLAGS) -S -emit-llvm $< -o $@

$(BUILDDIR)/%.o : $(SRCDIR)/%.cc
	$(CXX) $(CXXFLAGS) -c $< -o $@

$(BUILDDIR)/%.o : $(SRCDIR)/%.cpp
	$(CXX) $(CXXFLAGS) -c $< -o $@

$(BUILDDIR)/%.o : $(SRCDIR)/%.cpp
	$(CXX) $(CXXFLAGS) -c $< -o $@

no_opencl:
	@echo "OPENCL_PATH is not set. Open $(PARBOIL_ROOT)/common/Makefile.conf to set default value."
	@echo "You may use $(PLATFORM_MK) if you want a platform specific configurations."
	@exit 1

